# yapf: disable
# isort: skip_file
# ruff: noqa

# Standard imports.
from __future__ import annotations

import asyncio
import google.protobuf.descriptor
import google.protobuf.json_format
import google.protobuf.message
import grpc
import json
import logging
import sys
import traceback
import uuid
import resemble.aio.auth
import resemble.aio.auth.authorizers
import resemble.aio.auth.token_verifiers
import resemble.aio.aborted
import resemble.aio.state_managers
import resemble.v1alpha1.errors_pb2
import respect.logging  # type: ignore[import]
import typing
from abc import abstractmethod
from backoff import Backoff
from datetime import datetime, timedelta
from google.rpc import status_pb2
from grpc_status import rpc_status
from resemble.aio.contexts import (
    Context,
    ReaderContext,
    TransactionContext,
    WriterContext,
)
from resemble.aio.headers import Headers
from resemble.aio.idempotency import IdempotencyManager, Idempotency
from resemble.aio.internals.channel_manager import _ChannelManager
from resemble.aio.internals.middleware import Middleware
from resemble.aio.internals.tasks_cache import TasksCache
from resemble.aio.internals.tasks_dispatcher import TasksDispatcher
from resemble.aio.headers import Headers
from resemble.aio.servicers import Servicer, Serviceable
from resemble.aio.state_managers import Effects, StateManager
from resemble.aio.stubs import Stub
from resemble.aio.tasks import TaskEffect
from resemble.aio.types import (
    ActorId,
    ApplicationId,
    GrpcMetadata,
    ServiceName,
    assert_type,
)
from resemble.aio.workflows import Workflow
from resemble.v1alpha1.errors_pb2 import PermissionDenied
from typing import (
    cast,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    NoReturn,
    Optional,
    TypeAlias,
    Union,
)

# User defined or referenced imports.
import google.protobuf.empty_pb2
import resemble.cloud.v1alpha1.secrets.secrets_pb2
import resemble.cloud.v1alpha1.secrets.secrets_pb2_grpc
import resemble.v1alpha1.options_pb2
import resemble.v1alpha1.tasks_pb2
# Additionally re-export all messages from the pb2 module.
from resemble.cloud.v1alpha1.secrets.secrets_pb2 import (
    DeleteResponse,
    DeleteTaskResponse,
    ReadResponse,
    WriteRequest,
    WriteResponse,
    WriteTaskResponse,
)

logger = respect.logging.get_logger(__name__)
# TODO(benh): for now, default to `WARNING`: we have warnings in this
# file that we expect users to want to see.
logger.setLevel(logging.WARNING)


def MakeLegacyGrpcServiceable(
    # A legacy gRPC servicer type can't be more specific than `type`,
    # because legacy gRPC servicers (as generated by the gRPC `protoc`
    # plugin) do not share any common base class other than `object`.
    servicer_type: type
) -> Serviceable:
    raise ValueError(f"Unknown legacy gRPC servicer type '{servicer_type}'")


class SecretServicerMiddleware(Middleware):

    def __init__(
        self,
        *,
        servicer: SecretServicer,
        application_id: ApplicationId,
        state_manager: StateManager,
        channel_manager: _ChannelManager,
        tasks_cache: TasksCache,
        token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier],
    ):
        super().__init__(
            application_id=application_id,
            service_name='resemble.cloud.v1alpha1.secrets.Secret',
            channel_manager=channel_manager,
        )

        self._servicer = servicer
        self._state_manager = state_manager
        self.tasks_dispatcher = TasksDispatcher(self.dispatch, tasks_cache)

        # Store the type of each method's request so that stored requests can be
        # deserialized into the correct type.
        self.request_type_by_method_name: dict[str, type[google.protobuf.message.Message]] = {
            'Write': resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
            'WriteTask': resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
            'Read': google.protobuf.empty_pb2.Empty,
            'Delete': google.protobuf.empty_pb2.Empty,
            'DeleteTask': google.protobuf.empty_pb2.Empty,
        }

        # Create authorizer.
        self._authorizer: Optional[resemble.aio.auth.authorizers.Authorizer] = servicer.authorizer()

        # Create token verifier.
        self._token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier] = (
            servicer.token_verifier() or token_verifier
        )

        if self._token_verifier is not None and self._authorizer is None:
            raise RuntimeError(
                f"'TokenVerifier' provided but no 'Authorizer' returned for "
                f"'resemble.cloud.v1alpha1.secrets.Secret'"
            )


    def add_to_server(self, server: grpc.aio.Server) -> None:
        resemble.cloud.v1alpha1.secrets.secrets_pb2_grpc.add_SecretServicer_to_server(
            self, server
        )

    async def inspect(self, actor_id: ActorId) -> AsyncIterator[google.protobuf.message.Message]:
        """Implementation of `Middleware.inspect()`."""
        context = self.create_context(
            headers=Headers(
                application_id=self.application_id,
                service_name=self.service_name,
                actor_id=actor_id,
            ),
            context_type=ReaderContext,
        )

        async with self._state_manager.streaming_reader_idempotency_key(
            context, self._servicer.__state_type__
        ) as states:
            async for (state, idempotency_key) in states:
                yield state

    async def react_query(
        self,
        grpc_context: grpc.aio.ServicerContext,
        headers: Headers,
        method: str,
        request_bytes: bytes,
    ) -> AsyncIterator[tuple[Optional[google.protobuf.message.Message], list[uuid.UUID]]]:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes' for each state
        update that creates a different response.

        NOTE: only unary reader methods are supported."""
        # Need to define these up here since we can only do that once.
        last_response: Optional[google.protobuf.message.Message] = None
        aggregated_idempotency_keys: list[uuid.UUID] = []
        if 'Write' == method:
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            raise ValueError(f"Method '{method}' is invalid")
            yield  # Necessary for type checking.
        elif 'WriteTask' == method:
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            raise ValueError(f"Method '{method}' is invalid")
            yield  # Necessary for type checking.
        elif 'Read' == method:
            try:
                context = self.create_context(
                    headers=headers,
                    context_type=ReaderContext,
                )

                context.auth = await self._maybe_verify_token(headers)

                request = google.protobuf.empty_pb2.Empty()
                request.ParseFromString(request_bytes)

                async with self._state_manager.reactively(
                    context, self._servicer.__state_type__
                ) as states:
                    async for (state, idempotency_keys) in states:
                        # We currently authorize for each new `state`
                        # as a change in the state might now raise a
                        # `PermissionDenied`.
                        await self._maybe_authorize(
                            method_name=
                            'resemble.cloud.v1alpha1.secrets.Secret.Read',
                            headers=headers,
                            auth=context.auth,
                            state=state,
                            request=request,
                        )

                        aggregated_idempotency_keys.extend(idempotency_keys)

                        response = await self.__Read(
                            context, state, request
                        )

                        if last_response != response:
                            yield (response, aggregated_idempotency_keys)
                            last_response = response
                        else:
                            yield (None, aggregated_idempotency_keys)

                        aggregated_idempotency_keys.clear()
            except resemble.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    rpc_status.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
        elif 'Delete' == method:
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            raise ValueError(f"Method '{method}' is invalid")
            yield  # Necessary for type checking.
        elif 'DeleteTask' == method:
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            raise ValueError(f"Method '{method}' is invalid")
            yield  # Necessary for type checking.
        else:
            raise RuntimeError(f"Method '{method}' not found")
            yield  # Necessary for type checking.

    async def react_mutate(
        self,
        headers: Headers,
        method: str,
        request_bytes: bytes,
    ) -> google.protobuf.message.Message:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes'."""
        if 'Write' == method:
            request = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._Write(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        )
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        elif 'WriteTask' == method:
            request = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._WriteTask(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        )
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        elif 'Read' == method:
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            raise ValueError(f"Method '{method}' is invalid")
        elif 'Delete' == method:
            request = google.protobuf.empty_pb2.Empty()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._Delete(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        )
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        elif 'DeleteTask' == method:
            request = google.protobuf.empty_pb2.Empty()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._DeleteTask(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        )
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        else:
            raise RuntimeError(f"Method '{method}' not found")

    async def dispatch(
        self, task: TaskEffect, *, only_validate: bool = False
    ) -> google.protobuf.message.Message:
        """Dispatches the tasks to execute unless 'only_validate' is set to
        true, in which case just ensures that the task actually exists.
        Note that this function will be called *by* tasks_dispatcher; it will
        not itself call into tasks_dispatcher."""
        # Need to forward declare context since we set it to different
        # types depending on if the task is a reader or a writer.
        context: Context

        if 'WriteTask' == task.method_name:
            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def _run_WriteTask_task():
                if only_validate:
                    # TODO(benh): validate 'task.request' is correct type.
                    return True

                context = self.create_context(
                    headers=Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        actor_id=task.task_id.actor_id
                    ),
                    context_type=WriterContext,
                )

                async with self._state_manager.task_writer(
                    context, self._servicer.__state_type__
                ) as (state, task_writer):
                    effects = await self.__WriteTask(
                        context, state, task.request
                    )

                    if effects.tasks is not None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                    await task_writer.complete(task, effects)

                    if effects.tasks is not None:
                        self.tasks_dispatcher.dispatch(effects.tasks)

                    return effects.response


            return await _run_WriteTask_task()

        elif 'DeleteTask' == task.method_name:
            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            async def _run_DeleteTask_task():
                if only_validate:
                    # TODO(benh): validate 'task.request' is correct type.
                    return True

                context = self.create_context(
                    headers=Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        actor_id=task.task_id.actor_id
                    ),
                    context_type=WriterContext,
                )

                async with self._state_manager.task_writer(
                    context, self._servicer.__state_type__
                ) as (state, task_writer):
                    effects = await self.__DeleteTask(
                        context, state, task.request
                    )

                    if effects.tasks is not None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                    await task_writer.complete(task, effects)

                    if effects.tasks is not None:
                        self.tasks_dispatcher.dispatch(effects.tasks)

                    return effects.response


            return await _run_DeleteTask_task()


        # There are no tasks for this service.
        start_or_validate = "start" if not only_validate else "validate"
        raise RuntimeError(
            f"Attempted to {start_or_validate} task '{task.method_name}' "
            f"on 'Secret' which does not exist"
        )

    # Secret specific methods:
    async def __Write(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> Secret.WriteEffects:
        try:
            return await self._servicer.Write(
                context=context,
                state=state,
                request=request
            )
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            aborted_type: Optional[type] = None
            aborted_type = Secret.WriteAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                logger.warning(
                    f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {aborted}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(aborted))
                )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)  # type: ignore[attr-defined]
                ):
                    logger.warning(
                        f"Propagating unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {aborted}"
                    )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    logger.warning(
                        f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            logger.warning(
                f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                ''.join(traceback.format_exception(exception))
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Write') {type(exception).__name__}: {exception}"
            )


    async def _Write(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        context: WriterContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        response_bytes: Optional[bytes] = (
            self._state_manager.get_response_if_idempotent_mutation(
                context
            )
        )

        if response_bytes is not None:
            response = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse()
            response.ParseFromString(response_bytes)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )

            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.cloud.v1alpha1.secrets.Secret.Write',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__Write(
                    context,
                    state,
                    request
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def Write(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        context: Optional[WriterContext] = None
        try:
            context = self.create_context_from_grpc(
                grpc_context=grpc_context, context_type=WriterContext
            )

            assert context is not None

            return await self._Write(
                request,
                context,
            )
        except resemble.aio.aborted.Aborted as aborted:
            await grpc_context.abort_with_status(
                rpc_status.to_status(aborted.to_status())
            )
            raise  # Unreachable but necessary for mypy.
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except:
            # Print the exception stack trace for easier debugging. Note
            # that we don't include the stack trace in an error message
            # for the same reason that gRPC doesn't do so by default,
            # see https://github.com/grpc/grpc/issues/14897, but since this
            # should only get logged on the server side it is safe.
            logger.warning(
                'Unhandled exception\n' +
                ''.join(traceback.format_exc())
            )

            # Re-raise the exception for gRPC to handle!
            raise
        finally:
            if context is not None and context.transaction_id is not None:
                # Propagate transaction participants.
                #
                # NOTE: we'll want to propagate participants if we're
                # in a transaction method once we support nested
                # transactions, but for now we only need to propagate
                # participants when we're not in a transaction method.
                if not isinstance(context, TransactionContext):
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

    async def __WriteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> Secret.WriteTaskEffects:
        try:
            return await self._servicer.WriteTask(
                context=context,
                state=state,
                request=request
            )
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            aborted_type: Optional[type] = None
            aborted_type = Secret.WriteTaskAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                logger.warning(
                    f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {aborted}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(aborted))
                )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)  # type: ignore[attr-defined]
                ):
                    logger.warning(
                        f"Propagating unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {aborted}"
                    )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    logger.warning(
                        f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            logger.warning(
                f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                ''.join(traceback.format_exception(exception))
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask') {type(exception).__name__}: {exception}"
            )


    async def _WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        context: WriterContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        response_bytes: Optional[bytes] = (
            self._state_manager.get_response_if_idempotent_mutation(
                context
            )
        )

        if response_bytes is not None:
            response = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse()
            response.ParseFromString(response_bytes)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )

            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.cloud.v1alpha1.secrets.Secret.WriteTask',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__WriteTask(
                    context,
                    state,
                    request
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        raise RuntimeError('Tasks are not (yet) callable directly')

    async def __Read(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        try:
            return await self._servicer.Read(
                context=context,
                state=state,
                request=request
            )
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            aborted_type: Optional[type] = None
            aborted_type = Secret.ReadAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                logger.warning(
                    f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {aborted}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(aborted))
                )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)  # type: ignore[attr-defined]
                ):
                    logger.warning(
                        f"Propagating unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {aborted}"
                    )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    logger.warning(
                        f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            logger.warning(
                f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                ''.join(traceback.format_exception(exception))
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Read') {type(exception).__name__}: {exception}"
            )


    async def _Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: ReaderContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )

            async with self._state_manager.reader(
                context, self._servicer.__state_type__
            ) as state:
                await self._maybe_authorize(
                    method_name=
                    'resemble.cloud.v1alpha1.secrets.Secret.Read',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                response = await self.__Read(
                    context,
                    state,
                    request
                )
                if transaction is not None:
                    # We need the transaction to be stored _before_
                    # returning a response to the user otherwise they
                    # may read two different states if we crash.
                    await self._state_manager.transaction_participant_store(
                        transaction
                    )
                return response

    async def Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        context: Optional[ReaderContext] = None
        try:
            context = self.create_context_from_grpc(
                grpc_context=grpc_context, context_type=ReaderContext
            )

            assert context is not None

            return await self._Read(
                request,
                context,
            )
        except resemble.aio.aborted.Aborted as aborted:
            await grpc_context.abort_with_status(
                rpc_status.to_status(aborted.to_status())
            )
            raise  # Unreachable but necessary for mypy.
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except:
            # Print the exception stack trace for easier debugging. Note
            # that we don't include the stack trace in an error message
            # for the same reason that gRPC doesn't do so by default,
            # see https://github.com/grpc/grpc/issues/14897, but since this
            # should only get logged on the server side it is safe.
            logger.warning(
                'Unhandled exception\n' +
                ''.join(traceback.format_exc())
            )

            # Re-raise the exception for gRPC to handle!
            raise
        finally:
            if context is not None and context.transaction_id is not None:
                # Propagate transaction participants.
                #
                # NOTE: we'll want to propagate participants if we're
                # in a transaction method once we support nested
                # transactions, but for now we only need to propagate
                # participants when we're not in a transaction method.
                if not isinstance(context, TransactionContext):
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

    async def __Delete(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> Secret.DeleteEffects:
        try:
            return await self._servicer.Delete(
                context=context,
                state=state,
                request=request
            )
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            aborted_type: Optional[type] = None
            aborted_type = Secret.DeleteAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                logger.warning(
                    f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {aborted}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(aborted))
                )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)  # type: ignore[attr-defined]
                ):
                    logger.warning(
                        f"Propagating unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {aborted}"
                    )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    logger.warning(
                        f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            logger.warning(
                f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                ''.join(traceback.format_exception(exception))
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.Delete') {type(exception).__name__}: {exception}"
            )


    async def _Delete(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: WriterContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        response_bytes: Optional[bytes] = (
            self._state_manager.get_response_if_idempotent_mutation(
                context
            )
        )

        if response_bytes is not None:
            response = resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse()
            response.ParseFromString(response_bytes)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )

            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.cloud.v1alpha1.secrets.Secret.Delete',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__Delete(
                    context,
                    state,
                    request
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def Delete(
        self,
        request: google.protobuf.empty_pb2.Empty,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        context: Optional[WriterContext] = None
        try:
            context = self.create_context_from_grpc(
                grpc_context=grpc_context, context_type=WriterContext
            )

            assert context is not None

            return await self._Delete(
                request,
                context,
            )
        except resemble.aio.aborted.Aborted as aborted:
            await grpc_context.abort_with_status(
                rpc_status.to_status(aborted.to_status())
            )
            raise  # Unreachable but necessary for mypy.
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except:
            # Print the exception stack trace for easier debugging. Note
            # that we don't include the stack trace in an error message
            # for the same reason that gRPC doesn't do so by default,
            # see https://github.com/grpc/grpc/issues/14897, but since this
            # should only get logged on the server side it is safe.
            logger.warning(
                'Unhandled exception\n' +
                ''.join(traceback.format_exc())
            )

            # Re-raise the exception for gRPC to handle!
            raise
        finally:
            if context is not None and context.transaction_id is not None:
                # Propagate transaction participants.
                #
                # NOTE: we'll want to propagate participants if we're
                # in a transaction method once we support nested
                # transactions, but for now we only need to propagate
                # participants when we're not in a transaction method.
                if not isinstance(context, TransactionContext):
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

    async def __DeleteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> Secret.DeleteTaskEffects:
        try:
            return await self._servicer.DeleteTask(
                context=context,
                state=state,
                request=request
            )
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            aborted_type: Optional[type] = None
            aborted_type = Secret.DeleteTaskAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                logger.warning(
                    f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {aborted}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(aborted))
                )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)  # type: ignore[attr-defined]
                ):
                    logger.warning(
                        f"Propagating unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {aborted}"
                    )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    logger.warning(
                        f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            logger.warning(
                f"Unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                ''.join(traceback.format_exception(exception))
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask') {type(exception).__name__}: {exception}"
            )


    async def _DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: WriterContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        response_bytes: Optional[bytes] = (
            self._state_manager.get_response_if_idempotent_mutation(
                context
            )
        )

        if response_bytes is not None:
            response = resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse()
            response.ParseFromString(response_bytes)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )

            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=False
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__DeleteTask(
                    context,
                    state,
                    request
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        raise RuntimeError('Tasks are not (yet) callable directly')


    async def _maybe_authorize(
        self,
        *,
        method_name: str,
        headers: Headers,
        auth: Optional[resemble.aio.auth.Auth],
        state: Optional[SecretStateType] = None,
        request: Optional[SecretRequestTypes] = None,
    ) -> None:
        """Check authorization decision for the given method, using available
        information.

        Raises `PermissionDenied` in case authorizer is present but the request
        is not authorized.
        """
        if self._authorizer is not None:
            if not self._authorizer.can_authorize(method_name):
                # Better safe than sorry.
                raise resemble.aio.aborted.SystemAborted(
                    PermissionDenied(),
                    message=
                        f"Authorizer does not support method '{method_name}'"
                )

            if state is not None:
                # TODO: If we can read-only a state object we
                # can avoid a copy here.
                state_copy = SecretStateType()
                state_copy.CopyFrom(state)
                state = state_copy

            context = self.create_context(
                headers=(
                    # Get headers suitable for doing authorization.
                    headers.copy_for_token_verification_and_authorization()
                ),
                context_type=ReaderContext,
            )

            context.auth = auth

            try:
                authorized = await self._authorizer.authorize(
                    method_name=method_name,
                    context=context,
                    state=state,
                    request=request,
                )
            except resemble.aio.aborted.SystemAborted as aborted:
                # TODO(benh): when we refactor `authorize()` to return
                # `Ok | Unauthenticated | PermissionDenied` don't
                # catch an exception just print the error as we do below.
                if isinstance(
                    aborted.error, resemble.v1alpha1.errors_pb2.Unauthenticated
                ):
                    logger.warning(
                        "Unauthenticated call to "
                        f"'{method_name}' on '{context.actor_id}'"
                    )
                raise
            else:
                if not authorized:
                    logger.warning(
                        "Permission denied call to "
                        f"'{method_name}' on '{context.actor_id}'"
                    )
                    raise resemble.aio.aborted.SystemAborted(
                        PermissionDenied(),
                        message=
                        f'You are not authorized to call {method_name} on '
                        f"'{context.actor_id}'"
                    )


    async def _maybe_verify_token(self, headers) -> Optional[resemble.aio.auth.Auth]:
        """Verify the bearer token and if a token verifier is present.

        Returns the (optional) `resemble.aio.auth.Auth` object
        produced by the token verifier if the token can be verified.
        """
        auth: Optional[resemble.aio.auth.Auth] = None

        if self._token_verifier is not None and headers.bearer_token is not None:

            auth = await self._token_verifier.verify_token(
                context=self.create_context(
                    headers=(
                        # Get headers suitable for doing token verification.
                        headers.copy_for_token_verification_and_authorization()
                    ),
                    context_type=ReaderContext,
                ),
                token=headers.bearer_token,
            )

        return auth


class _SecretStub(Stub):

    def __init__(
        self,
        *,
        context_or_workflow: Context | Workflow,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        context: Optional[Context] = None

        if isinstance(context_or_workflow, Context):
            context = context_or_workflow

            # ISSUE(https://github.com/reboot-dev/respect/issues/2258)
            # For now we do not explicitly pass on the bearer token here, as it
            # might have security implications to just pass on the token to any
            # service we are calling.

        elif isinstance(context_or_workflow, Workflow):
            bearer_token = context_or_workflow.bearer_token

        super().__init__(
            channel_manager=context_or_workflow.channel_manager,
            idempotency_manager=context_or_workflow,
            service_name='resemble.cloud.v1alpha1.secrets.Secret',
            actor_id=actor_id,
            context=context,
            bearer_token=bearer_token,
        )

        channel = self._channel_manager.get_channel_for(
            'resemble.cloud.v1alpha1.secrets.Secret', actor_id
        )
        self._stub = resemble.cloud.v1alpha1.secrets.secrets_pb2_grpc.SecretStub(channel)


class SecretReaderStub(_SecretStub):

    def __init__(
        self,
        context_or_workflow: ReaderContext | WriterContext | TransactionContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(
            context_or_workflow,
            [ReaderContext, WriterContext, TransactionContext, Workflow]
        )
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Secret specific methods:


    async def Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        async with self._call(
            'Read',
            self._stub.Read,
            request,
            unary=True,
            reader=True,
            response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse,
            aborted_type=Secret.ReadAborted,
            metadata=metadata,
        ) as call:
            assert isinstance(call, Awaitable)
            return await call




class SecretWriterStub(_SecretStub):

    def __init__(
        self,
        context_or_workflow: TransactionContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context_or_workflow, [TransactionContext, Workflow])
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Secret specific methods:
    async def Write(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Write',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Write',
                self._stub.Write,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse,
                aborted_type=Secret.WriteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='WriteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'WriteTask',
                self._stub.WriteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse,
                aborted_type=Secret.WriteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        async with self._call(
            'Read',
            self._stub.Read,
            request,
            unary=True,
            reader=True,
            response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse,
            aborted_type=Secret.ReadAborted,
            metadata=metadata,
        ) as call:
            assert isinstance(call, Awaitable)
            return await call

    async def Delete(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Delete',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Delete',
                self._stub.Delete,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse,
                aborted_type=Secret.DeleteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='DeleteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'DeleteTask',
                self._stub.DeleteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse,
                aborted_type=Secret.DeleteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call


class SecretTransactionStub(_SecretStub):

    def __init__(
        self,
        context_or_workflow: TransactionContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context_or_workflow, [TransactionContext, Workflow])
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Secret specific methods:
    async def Write(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Write',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Write',
                self._stub.Write,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse,
                aborted_type=Secret.WriteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='WriteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'WriteTask',
                self._stub.WriteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse,
                aborted_type=Secret.WriteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        async with self._call(
            'Read',
            self._stub.Read,
            request,
            unary=True,
            reader=True,
            response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse,
            aborted_type=Secret.ReadAborted,
            metadata=metadata,
        ) as call:
            assert isinstance(call, Awaitable)
            return await call

    async def Delete(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Delete',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Delete',
                self._stub.Delete,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse,
                aborted_type=Secret.DeleteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='DeleteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'DeleteTask',
                self._stub.DeleteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse,
                aborted_type=Secret.DeleteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call


class SecretWorkflowStub(_SecretStub):

    def __init__(
        self,
        *,
        workflow: Workflow,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(workflow, [Workflow])
        super().__init__(
            context_or_workflow=workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Secret specific methods:
    async def Write(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Write',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Write',
                self._stub.Write,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse,
                aborted_type=Secret.WriteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='WriteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'WriteTask',
                self._stub.WriteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse,
                aborted_type=Secret.WriteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def Read(
        self,
        request: google.protobuf.empty_pb2.Empty,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        async with self._call(
            'Read',
            self._stub.Read,
            request,
            unary=True,
            reader=True,
            response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse,
            aborted_type=Secret.ReadAborted,
            metadata=metadata,
        ) as call:
            assert isinstance(call, Awaitable)
            return await call

    async def Delete(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Delete',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Delete',
                self._stub.Delete,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse,
                aborted_type=Secret.DeleteAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='DeleteTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'DeleteTask',
                self._stub.DeleteTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse,
                aborted_type=Secret.DeleteTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call


class SecretTasksStub(_SecretStub):

    def __init__(
        self,
        context: WriterContext,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context, [WriterContext])
        super().__init__(
            context_or_workflow=context,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Secret specific methods:

    def WriteTask(
        self,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
        *,
        schedule: Optional[datetime | timedelta] = None,
    ) -> TaskEffect:
        return TaskEffect(
            service_name='resemble.cloud.v1alpha1.secrets.Secret',
            actor_id=self._headers.actor_id,
            method_name='WriteTask',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )



    def DeleteTask(
        self,
        request: google.protobuf.empty_pb2.Empty,
        *,
        schedule: Optional[datetime | timedelta] = None,
    ) -> TaskEffect:
        return TaskEffect(
            service_name='resemble.cloud.v1alpha1.secrets.Secret',
            actor_id=self._headers.actor_id,
            method_name='DeleteTask',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )



class SecretAuthorizerMixin:
    """Mixin that provides a default implementation of the `can_authorize`
    method for all Secret authorizers.
    """

    def can_authorize(self, method_name: str) -> bool:
        service_method_names = [
            'resemble.cloud.v1alpha1.secrets.Secret.Write',
            'resemble.cloud.v1alpha1.secrets.Secret.WriteTask',
            'resemble.cloud.v1alpha1.secrets.Secret.Read',
            'resemble.cloud.v1alpha1.secrets.Secret.Delete',
            'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask',
        ]
        return method_name in service_method_names


SecretStateType: TypeAlias = google.protobuf.empty_pb2.Empty
SecretRequestTypes: TypeAlias = \
        resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest \
        | resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest \
        | google.protobuf.empty_pb2.Empty \
        | google.protobuf.empty_pb2.Empty


class SecretAuthorizer(
    SecretAuthorizerMixin,
    resemble.aio.auth.authorizers.Authorizer[SecretStateType, SecretRequestTypes],
):
    StateType: TypeAlias = SecretStateType
    RequestTypes: TypeAlias = SecretRequestTypes

    async def authorize(
        self,
        *,
        method_name: str,
        context: ReaderContext,
        state: Optional[SecretStateType] = None,
        request: Optional[SecretRequestTypes] = None,
    ) -> bool:
        if method_name == 'resemble.cloud.v1alpha1.secrets.Secret.Write':
            return await self.Write(
                context,
                cast(google.protobuf.empty_pb2.Empty, state),
                cast(resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest, request),
            )
        elif method_name == 'resemble.cloud.v1alpha1.secrets.Secret.WriteTask':
            return await self.WriteTask(
                context,
                cast(google.protobuf.empty_pb2.Empty, state),
                cast(resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest, request),
            )
        elif method_name == 'resemble.cloud.v1alpha1.secrets.Secret.Read':
            return await self.Read(
                context,
                cast(google.protobuf.empty_pb2.Empty, state),
                cast(google.protobuf.empty_pb2.Empty, request),
            )
        elif method_name == 'resemble.cloud.v1alpha1.secrets.Secret.Delete':
            return await self.Delete(
                context,
                cast(google.protobuf.empty_pb2.Empty, state),
                cast(google.protobuf.empty_pb2.Empty, request),
            )
        elif method_name == 'resemble.cloud.v1alpha1.secrets.Secret.DeleteTask':
            return await self.DeleteTask(
                context,
                cast(google.protobuf.empty_pb2.Empty, state),
                cast(google.protobuf.empty_pb2.Empty, request),
            )
        else:
            return False

    async def Write(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> bool:
        return False

    async def WriteTask(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> bool:
        return False

    async def Read(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> bool:
        return False

    async def Delete(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> bool:
        return False

    async def DeleteTask(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> bool:
        return False


    # Off-the shelf authorizers for Secret:

    class AllowAllIfAuthenticated(
        SecretAuthorizerMixin,
        resemble.aio.auth.authorizers.Authorizer,
    ):
        """An authorizer for Secret that allows all requests if the
        caller is authenticated.
        """
        async def authorize(
            self,
            *,
            method_name: str,
            context: ReaderContext,
            state: Optional[SecretStateType] = None,
            request: Optional[SecretRequestTypes] = None,
        ) -> bool:
            return context.auth is not None


class SecretServicer(Servicer):
    Authorizer: TypeAlias = SecretAuthorizer

    __service_name__ = 'resemble.cloud.v1alpha1.secrets.Secret'
    __file_descriptor__ = resemble.cloud.v1alpha1.secrets.secrets_pb2.DESCRIPTOR
    __state_type__ = google.protobuf.empty_pb2.Empty

    def __init__(self):
        # NOTE: need to hold on to the middleware so we can do inline
        # writes (see 'self.write(...)').
        #
        # Because '_middleware' is not really private this does mean
        # users may do possibly dangerous things, but this is no more
        # likely given they could have already overridden
        # 'create_middleware()'.
        self._middleware: Optional[SecretServicerMiddleware] = None

    def create_middleware(
        self,
        *,
        application_id: ApplicationId,
        state_manager: StateManager,
        channel_manager: _ChannelManager,
        tasks_cache: TasksCache,
        token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier],
    ) -> SecretServicerMiddleware:
        self._middleware = SecretServicerMiddleware(
            servicer=self,
            application_id=application_id,
            state_manager=state_manager,
            channel_manager=channel_manager,
            tasks_cache=tasks_cache,
            token_verifier=token_verifier,
        )
        return self._middleware

    def authorizer(self) -> Optional[resemble.aio.auth.authorizers.Authorizer]:
        return None

    def token_verifier(self) -> Optional[resemble.aio.auth.token_verifiers.TokenVerifier]:
        return None

    class Tasks:
        def __init__(self, when: Optional[datetime | timedelta] = None):
            self._when = when

        # Secret specific methods:
        def WriteTask(
            self,
            rsm_context: WriterContext,
            *,
            data: Optional[bytes] = None,
        ) -> TaskEffect:
            if data is not None and not isinstance(
                data,
                bytes,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest': field 'data' is not "
                    f"of required type 'bytes'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments
            # are always non-None, when in reality they are optional.
            request = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest(
                data=data,  # type: ignore[arg-type]
            )
            return SecretTasksStub(
                context=rsm_context,
                actor_id=rsm_context.actor_id,
            ).WriteTask(
                request,
                schedule=self._when,
            )

        def DeleteTask(
            self,
            rsm_context: WriterContext,
        ) -> TaskEffect:
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments
            # are always non-None, when in reality they are optional.
            request = google.protobuf.empty_pb2.Empty(
            )
            return SecretTasksStub(
                context=rsm_context,
                actor_id=rsm_context.actor_id,
            ).DeleteTask(
                request,
                schedule=self._when,
            )


    def schedule(
        self,
        when: Optional[datetime | timedelta] = None,
    ):
        return self.Tasks(when)

    class Effects(resemble.aio.state_managers.Effects):
        def __init__(
            self,
            *,
            state: google.protobuf.empty_pb2.Empty,
            response: Optional[google.protobuf.message.Message] = None,
            tasks: Optional[list[TaskEffect]] = None
        ):
            assert_type(state, [google.protobuf.empty_pb2.Empty])

            super().__init__(state=state, response=response, tasks=tasks)

    class WriteEffects(Effects):
        def __init__(
            self,
            *,
            state: google.protobuf.empty_pb2.Empty,
            response: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse,
            tasks: Optional[list[TaskEffect]] = None
        ):
            assert_type(state, [google.protobuf.empty_pb2.Empty])
            assert_type(response, [resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse])

            super().__init__(state=state, response=response, tasks=tasks)


    class WriteTaskEffects(Effects):
        def __init__(
            self,
            *,
            state: google.protobuf.empty_pb2.Empty,
            response: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse,
            tasks: Optional[list[TaskEffect]] = None
        ):
            assert_type(state, [google.protobuf.empty_pb2.Empty])
            assert_type(response, [resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse])

            super().__init__(state=state, response=response, tasks=tasks)



    class DeleteEffects(Effects):
        def __init__(
            self,
            *,
            state: google.protobuf.empty_pb2.Empty,
            response: resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse,
            tasks: Optional[list[TaskEffect]] = None
        ):
            assert_type(state, [google.protobuf.empty_pb2.Empty])
            assert_type(response, [resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse])

            super().__init__(state=state, response=response, tasks=tasks)


    class DeleteTaskEffects(Effects):
        def __init__(
            self,
            *,
            state: google.protobuf.empty_pb2.Empty,
            response: resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse,
            tasks: Optional[list[TaskEffect]] = None
        ):
            assert_type(state, [google.protobuf.empty_pb2.Empty])
            assert_type(response, [resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse])

            super().__init__(state=state, response=response, tasks=tasks)



    class InlineWriterCallable(typing.Protocol):
        async def __call__(
            self,
            context: WriterContext,
            state: google.protobuf.empty_pb2.Empty
        ) -> SecretServicer.Effects:
            ...

    async def read(
        self, context: TransactionContext
    ) -> google.protobuf.empty_pb2.Empty:
        """Read the current state within a transaction."""
        assert_type(context, [TransactionContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        return await self._middleware._state_manager.read(
            context, self.__state_type__
        )

    async def write(
        self,
        context: TransactionContext,
        writer: InlineWriterCallable,
    ):
        """Perform an "inline write" within a transaction."""
        assert_type(context, [TransactionContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        writer_context: WriterContext = self._middleware.create_context(
            headers=Headers(
                application_id=context.application_id,
                service_name=self.__service_name__,
                actor_id=context.actor_id,
                transaction_id=context.transaction_id,
                transaction_coordinator_service=context
                .transaction_coordinator_service,
                transaction_coordinator_actor_id=context
                .transaction_coordinator_actor_id
            ),
            context_type=WriterContext,
        )

        async with self._middleware._state_manager.transactionally(
            writer_context, self._middleware.tasks_dispatcher
        ) as transaction:
            assert transaction is not None
            async with self._middleware._state_manager.writer(
                writer_context, self.__state_type__, transaction=transaction
            ) as (state, state_manager_writer):
                effects: SecretServicer.Effects = await writer(
                    context=writer_context, state=state
                )

                assert_type(
                    effects,
                    [SecretServicer.Effects],
                    may_be_subclass=False,
                )

                # NOTE: we validate tasks added as part of a
                # transaction when we prepare.

                await state_manager_writer.complete(effects)

                if effects.tasks is not None:
                    assert all(
                        task.task_id.service == transaction.service
                        for task in effects.tasks
                    ), 'Task service does not match transaction service'
                    transaction.tasks.extend(effects.tasks)

                return effects.response  # May be 'None'.

    # Secret specific methods:
    @abstractmethod
    async def Write(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> WriteEffects:
        raise NotImplementedError

    @abstractmethod
    async def WriteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> WriteTaskEffects:
        raise NotImplementedError

    @abstractmethod
    async def Read(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        raise NotImplementedError

    @abstractmethod
    async def Delete(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> DeleteEffects:
        raise NotImplementedError

    @abstractmethod
    async def DeleteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> DeleteTaskEffects:
        raise NotImplementedError


class Secret:

    Interface: TypeAlias = SecretServicer

    Effects: TypeAlias = SecretServicer.Effects

    Authorizer: TypeAlias = SecretAuthorizer

    WriteEffects: TypeAlias = Interface.WriteEffects

    class WriteAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False

    WriteTaskEffects: TypeAlias = Interface.WriteTaskEffects

    class WriteTaskAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False


    class ReadAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False

    DeleteEffects: TypeAlias = Interface.DeleteEffects

    class DeleteAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False

    DeleteTaskEffects: TypeAlias = Interface.DeleteTaskEffects

    class DeleteTaskAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False


    def __init__(
        self,
        actor_id: str,
        *,
        bearer_token: Optional[str] = None,
    ):
        self._actor_id = actor_id
        self._reader_stub: Optional[SecretReaderStub] = None
        self._writer_stub: Optional[SecretWriterStub] = None
        self._workflow_stub: Optional[SecretWorkflowStub] = None
        self._bearer_token = bearer_token

    def reader(
        self,
        context: ReaderContext | WriterContext | TransactionContext | Workflow,
    ) -> SecretReaderStub:
        if self._reader_stub is None:
            self._reader_stub = SecretReaderStub(
                context_or_workflow=context,
                actor_id=self._actor_id,
                bearer_token=self._bearer_token,
            )
        assert self._reader_stub is not None
        assert self._reader_stub._idempotency_manager == context
        return self._reader_stub

    def writer(
        self,
        context: TransactionContext | Workflow,
    ) -> SecretWriterStub:
        if self._writer_stub is None:
            self._writer_stub = SecretWriterStub(
                context_or_workflow=context,
                actor_id=self._actor_id,
                bearer_token=self._bearer_token,
            )
        assert self._writer_stub is not None
        assert self._writer_stub._idempotency_manager == context
        return self._writer_stub

    def workflow(
        self,
        workflow: Workflow,
    ) -> SecretWorkflowStub:
        if self._workflow_stub is None:
            self._workflow_stub = SecretWorkflowStub(
                workflow=workflow,
                actor_id=self._actor_id,
                bearer_token=self._bearer_token,
            )
        assert self._workflow_stub is not None
        assert self._workflow_stub._idempotency_manager == workflow
        return self._workflow_stub

    # Secret specific methods:
    async def Write(
        self,
        rsm_context_or_workflow: TransactionContext | Workflow,
        rsm_idempotency: Optional[Idempotency] = None,
        rsm_metadata: Optional[GrpcMetadata] = None,
        *,
        data: Optional[bytes] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse:
        if data is not None and not isinstance(
            data,
            bytes,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest': field 'data' is not "
                f"of required type 'bytes'"
            )
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest(
            data=data,  # type: ignore[arg-type]
        )
        return await self.writer(rsm_context_or_workflow).Write(
            request,
            rsm_idempotency,
            metadata=rsm_metadata,
        )

    async def WriteTask(
        self,
        rsm_context_or_workflow: TransactionContext | Workflow,
        rsm_idempotency: Optional[Idempotency] = None,
        rsm_metadata: Optional[GrpcMetadata] = None,
        *,
        data: Optional[bytes] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse:
        if data is not None and not isinstance(
            data,
            bytes,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest': field 'data' is not "
                f"of required type 'bytes'"
            )
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest(
            data=data,  # type: ignore[arg-type]
        )
        return await self.writer(rsm_context_or_workflow).WriteTask(
            request,
            rsm_idempotency,
            metadata=rsm_metadata,
        )

    async def Read(
        self,
        rsm_context_or_workflow: ReaderContext | WriterContext | TransactionContext | Workflow,
        rsm_metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = google.protobuf.empty_pb2.Empty(
        )
        return await self.reader(rsm_context_or_workflow).Read(
            request,
            metadata=rsm_metadata,
        )

    async def Delete(
        self,
        rsm_context_or_workflow: TransactionContext | Workflow,
        rsm_idempotency: Optional[Idempotency] = None,
        rsm_metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse:
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = google.protobuf.empty_pb2.Empty(
        )
        return await self.writer(rsm_context_or_workflow).Delete(
            request,
            rsm_idempotency,
            metadata=rsm_metadata,
        )

    async def DeleteTask(
        self,
        rsm_context_or_workflow: TransactionContext | Workflow,
        rsm_idempotency: Optional[Idempotency] = None,
        rsm_metadata: Optional[GrpcMetadata] = None,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse:
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = google.protobuf.empty_pb2.Empty(
        )
        return await self.writer(rsm_context_or_workflow).DeleteTask(
            request,
            rsm_idempotency,
            metadata=rsm_metadata,
        )


class SecretServicerNodeAdaptor(Secret.Interface):

    def __init__(self):
        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        self._js_servicer_reference = self._construct_js_servicer()  # type: ignore[attr-defined]

    # Secret specific methods:
    async def Write(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> Secret.WriteEffects:
        json_state = google.protobuf.json_format.MessageToJson(state)

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'Write',
            context,
            json_state,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Secret.WriteEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                google.protobuf.empty_pb2.Empty(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteResponse(),
            ),
        )

    async def WriteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteRequest,
    ) -> Secret.WriteTaskEffects:
        json_state = google.protobuf.json_format.MessageToJson(state)

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'WriteTask',
            context,
            json_state,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Secret.WriteTaskEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                google.protobuf.empty_pb2.Empty(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.cloud.v1alpha1.secrets.secrets_pb2.WriteTaskResponse(),
            ),
        )

    async def Read(
        self,
        context: ReaderContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse:
        json_state = google.protobuf.json_format.MessageToJson(state)
        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_response = await self._reader(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'Read',
            context,
            json_state,
            json_request,
        )

        return google.protobuf.json_format.Parse(
            json_response,
            resemble.cloud.v1alpha1.secrets.secrets_pb2.ReadResponse(),
        )

    async def Delete(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> Secret.DeleteEffects:
        json_state = google.protobuf.json_format.MessageToJson(state)

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'Delete',
            context,
            json_state,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Secret.DeleteEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                google.protobuf.empty_pb2.Empty(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteResponse(),
            ),
        )

    async def DeleteTask(
        self,
        context: WriterContext,
        state: google.protobuf.empty_pb2.Empty,
        request: google.protobuf.empty_pb2.Empty,
    ) -> Secret.DeleteTaskEffects:
        json_state = google.protobuf.json_format.MessageToJson(state)

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'DeleteTask',
            context,
            json_state,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Secret.DeleteTaskEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                google.protobuf.empty_pb2.Empty(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.cloud.v1alpha1.secrets.secrets_pb2.DeleteTaskResponse(),
            ),
        )



class SecretNodeAdaptor(Secret):

    async def _reader(
        self,
        method: str,
        context: ReaderContext | WriterContext | TransactionContext | Workflow,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        response = await getattr(super().reader(context), method)(request)

        return google.protobuf.json_format.MessageToJson(response)

    async def _writer(
        self,
        method: str,
        context: TransactionContext | Workflow,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        response = await getattr(super().writer(context), method)(request)

        return google.protobuf.json_format.MessageToJson(response)

# yapf: enable
